import gzip
import os
import shutil
import tarfile
import zipfile
import requests
import logging

logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class DownloadDataset:
    r"""Download class for datasets. Provides static-methods and functions for download and unzip of the data.
    They are intentionally kept general and could also be used outside of this class definition.
    Dataset-specific functions like :obj:`prepare_data` must be implemented in subclasses.
    Note that :obj:``DownloadDataset`` uses a main directory located at '~/.kgcnn/datasets' for downloading datasets
    as default. Classes in :obj:`kgcnn.data.datasets` inherit from this class, but :obj:`DownloadDataset` can also be
    used as a member.
    """

    def __init__(self,
                 dataset_name: str = None,
                 download_file_name: str = None,
                 data_directory_name: str = None,
                 unpack_directory_name: str = None,
                 extract_file_name: str = None,
                 download_url: str = None,
                 unpack_tar: bool = False,
                 unpack_zip: bool = False,
                 extract_gz: bool = False,
                 reload: bool = False,
                 verbose: int = 10,
                 data_main_dir: str = os.path.join(os.path.expanduser("~"), ".kgcnn", "datasets")
                 ):
        r"""Base initialization function for downloading and extracting the data. The arguments to the constructor
        determine what to download and whether to unpack the download. The main function :obj:`download_dataset_to_disk`
        is already called in the constructor.

        Args:
            dataset_name (str): Name of the dataset to download (optional). Default is None.
            download_file_name (str): Name of the file that the source is downloaded to. Default is None.
            data_directory_name (str): Name of the data directory in data_main_dir the file is saved. Default is None.
            unpack_directory_name (str): The name of a new directory in data_directory_name to unpack archive.
                Default is None.
            extract_file_name (str): Name of a gz-file to extract. Default is None.
            download_url (str): Url for file to download. Default is None.
            unpack_tar (bool): Whether to unpack a tar-archive. Default is False.
            unpack_zip (bool): Whether to unpack a zip-archive. Default is False.
            extract_gz (bool): Whether to unpack a gz-archive. Default is False.
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Logging level. Default is 10.
        """
        self.data_main_dir = data_main_dir
        self.dataset_name = dataset_name
        self.download_file_name = download_file_name
        self.data_directory_name = data_directory_name
        self.unpack_directory_name = unpack_directory_name
        self.extract_file_name = extract_file_name
        self.download_url = download_url
        self.unpack_tar = unpack_tar
        self.unpack_zip = unpack_zip
        self.extract_gz = extract_gz
        self.download_reload = reload
        self.logger_download = module_logger
        self.logger_download.setLevel(verbose)

        # Make the download already in init.
        self.download_dataset_to_disk()

    def download_dataset_to_disk(self):
        """Main download function to run the download and unpack of the dataset. Defined by attributes in self."""

        # Some datasets do not offer all information or require multiple files.
        self.logger_download.info("Checking and possibly downloading dataset with name %s" % str(self.dataset_name))

        # Default functions to load a dataset.
        if self.data_directory_name is not None:
            self.setup_dataset_main(self.data_main_dir, logger=self.logger_download)
            self.setup_dataset_dir(self.data_main_dir, self.data_directory_name, logger=self.logger_download)

        if self.download_url is not None:
            self.download_database(os.path.join(self.data_main_dir, self.data_directory_name), self.download_url,
                                   self.download_file_name, overwrite=self.download_reload, logger=self.logger_download)

        if self.unpack_tar:
            self.unpack_tar_file(os.path.join(self.data_main_dir, self.data_directory_name), self.download_file_name,
                                 self.unpack_directory_name, overwrite=self.download_reload,
                                 logger=self.logger_download)

        if self.unpack_zip:
            self.unpack_zip_file(os.path.join(self.data_main_dir, self.data_directory_name), self.download_file_name,
                                 self.unpack_directory_name, overwrite=self.download_reload,
                                 logger=self.logger_download)

        if self.extract_gz:
            self.extract_gz_file(os.path.join(self.data_main_dir, self.data_directory_name), self.download_file_name,
                                 self.extract_file_name, overwrite=self.download_reload, logger=self.logger_download)

    @staticmethod
    def setup_dataset_main(data_main_dir, logger=None):
        """Make the main-directory for all datasets to store data.

        Args:
           data_main_dir (str): Path to create directory.
           logger: Logger to print information or warnings.
        """
        os.makedirs(data_main_dir, exist_ok=True)
        if logger is not None:
            logger.info("Dataset directory located at %s" % data_main_dir)

    @staticmethod
    def setup_dataset_dir(data_main_dir: str, data_directory: str, logger=None):
        """Make directory for each dataset.

        Args:
            data_main_dir (str): Path-location of the directory for all datasets.
            data_directory (str): Path of the directory for specific dataset to create.
            logger: Logger to print information or warnings.
        """
        if not os.path.exists(os.path.join(data_main_dir, data_directory)):
            os.mkdir(os.path.join(data_main_dir, data_directory))
        else:
            if logger is not None:
                logger.info("Dataset directory found. Done.")

    @staticmethod
    def download_database(path: str, download_url: str, filename: str, overwrite: bool = False, logger=None):
        """Download dataset file.

        Args:
            path (str): Target filepath to store file (without filename).
            download_url (str): String of the download url to catch database from.
            filename (str): Name the dataset is downloaded to.
            overwrite (bool): Overwrite existing database. Default is False.
            logger: Logger to print information or warnings.

        Returns:
            os.path: Filepath of downloaded file.
        """
        if os.path.exists(os.path.join(path, filename)) is False or overwrite:
            if logger is not None:
                logger.info("Downloading dataset... ")
            r = requests.get(download_url, allow_redirects=True)
            open(os.path.join(path, filename), 'wb').write(r.content)
        else:
            if logger is not None:
                logger.info("Dataset found. Done.")
        return os.path.join(path, filename)

    @staticmethod
    def unpack_tar_file(path: str, filename: str, unpack_directory: str, overwrite: bool = False, logger=None):
        """Extract tar-file.

        Args:
            path (str): Filepath where the tar-file to extract is located.
            filename (str): Name of the dataset to extract.
            unpack_directory (str): Directory to extract data to.
            overwrite (bool): Overwrite existing database. Default is False.
            logger: Logger to print information or warnings.

        Returns:
            os.path: Filepath of the extracted dataset folder.
        """
        if not os.path.exists(os.path.join(path, unpack_directory)):
            if logger is not None:
                logger.info("Creating directory... ")
            os.mkdir(os.path.join(path, unpack_directory))
        else:
            if logger is not None:
                logger.info("Directory for extraction exists. Done.")
            if not overwrite:
                if logger is not None:
                    logger.info("Not extracting tar File. Stopped.")
                return os.path.join(path, unpack_directory)  # Stop extracting here

        if logger is not None:
            logger.info("Read tar file... ")
        archive = tarfile.open(os.path.join(path, filename), "r")
        # Filelistnames = archive.getnames()

        if logger is not None:
            logger.info("Extracting tar file... ")
        archive.extractall(os.path.join(path, unpack_directory))
        archive.close()
        return os.path.join(path, unpack_directory)

    @staticmethod
    def unpack_zip_file(path: str, filename: str, unpack_directory: str, overwrite: bool = False, logger=None):
        """Extract zip-file.

        Args:
            path (str): Filepath where the zip-file to extract is located.
            filename (str): Name of the dataset to extract.
            unpack_directory (str): Directory to extract data to.
            overwrite (bool): Overwrite existing database. Default is False.
            logger: Logger to print information or warnings.

        Returns:
            os.path: Filepath of the extracted dataset folder.
        """
        if os.path.exists(os.path.join(path, unpack_directory)):
            if logger is not None:
                logger.info("Directory for extraction exists. Done.")
            if not overwrite:
                if logger is not None:
                    logger.info("Not extracting zip file. Stopped.")
                return os.path.join(path, unpack_directory)

        if logger is not None:
            logger.info("Read zip file ... ")
        archive = zipfile.ZipFile(os.path.join(path, filename), "r")
        # Filelistnames = archive.getnames()
        if logger is not None:
            logger.info("Extracting zip file...")
        archive.extractall(os.path.join(path, unpack_directory))
        archive.close()
        return os.path.join(path, unpack_directory)

    @staticmethod
    def extract_gz_file(path: str, filename: str, out_filename: str = None,
                        overwrite: bool = False, logger=None):
        """Extract gz-file.

        Args:
            path (str): Filepath where the gz-file to extract is located.
            filename (str): Name of the gz-file to extract.
            out_filename (str): Name of the extracted file.
            overwrite (bool): Overwrite existing database. Default is False.
            logger: Logger to print information or warnings.

        Returns:
            os.path: Filepath of the extracted file.
        """
        if out_filename is None:
            out_filename = filename.replace(".gz", "")

        if os.path.exists(os.path.join(path, out_filename)):
            if logger is not None:
                logger.info("Extracted file exists. Done.")
            if not overwrite:
                if logger is not None:
                    logger.info("Not extracting gz-file. Stopped.")
                return os.path.join(path, out_filename)

        if logger is not None:
            logger.info("Extract gz-file ... ")

        with gzip.open(os.path.join(path, filename), 'rb') as f_in:
            with open(os.path.join(path, out_filename), 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)

        return os.path.join(path, out_filename)



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.ops.partition import change_partition_by_name, partition_row_indexing
from kgcnn.layers.base import GraphBaseLayer


@tf.keras.utils.register_keras_serializable(package='kgcnn',name='PoolingTopK')
class PoolingTopK(GraphBaseLayer):
    """Layer for pooling of nodes. Disjoint representation including length tensor.
    
    This implements a learnable score vector plus gate. Implements gPool of Gao et al.
    
    Args:
        k (float): relative number of nodes to remove. Default is 0.1
        kernel_initializer (str): Score initialization. Default is 'glorot_uniform',
        kernel_regularizer (str): Score regularization. Default is None.
        kernel_constraint (bool): Score constrain. Default is None.
    """

    def __init__(self,
                 k=0.1,
                 kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 kernel_constraint=None,
                 **kwargs):
        """Initialize Layer."""
        super(PoolingTopK, self).__init__(**kwargs)
        self.k = k
        self.kernel_initializer = ks.initializers.get(kernel_initializer)
        self.kernel_regularizer = ks.regularizers.get(kernel_regularizer)
        self.kernel_constraint = ks.constraints.get(kernel_constraint)

        self.units_p = None
        self.kernel_p = None

    def build(self, input_shape):
        """Build Layer."""
        super(PoolingTopK, self).build(input_shape)
        self.units_p = int(input_shape[0][-1])
        self.kernel_p = self.add_weight('score',
                                        shape=[1, self.units_p],
                                        initializer=self.kernel_initializer,
                                        regularizer=self.kernel_regularizer,
                                        constraint=self.kernel_constraint,
                                        dtype=self.dtype,
                                        trainable=True)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, node_partition, edges, edge_partition, edge_indices]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Edge embeddings of shape (batch, [N], F)
                - edge_indices (tf.RaggedTensor): Edge index list of shape of shape (batch, [N], 2)
        
        Returns:
            tuple: [nodes, edges, edge_indices], [map_nodes, map_edges]
            
                - nodes (tf.RaggedTensor): Pooled node feature tensor
                - edges (tf.RaggedTensor): Pooled edge feature list
                - edge_indices (tf.RaggedTensor): Pooled edge index list
                - map_nodes (tf.RaggedTensor): Index map between original and pooled nodes
                - map_edges (tf.RaggedTensor): Index map between original and pooled edges
        """
        self.assert_ragged_input_rank(inputs)
        dyn_inputs = inputs
        # We cast to values here
        node, nodelen = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
        edgefeat, edgelen = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
        edgeindref, _ = dyn_inputs[2].values, dyn_inputs[2].row_lengths()

        index_dtype = edgeindref.dtype
        # Get node properties
        nvalue = node
        nrowlength = tf.cast(nodelen, dtype=index_dtype)
        erowlength = tf.cast(edgelen, dtype=index_dtype)
        nids = tf.repeat(tf.range(tf.shape(nrowlength)[0], dtype=index_dtype), nrowlength)

        # Use kernel p to get score
        norm_p = ks.backend.sqrt(ks.backend.sum(ks.backend.square(self.kernel_p), axis=-1, keepdims=True))
        nscore = ks.backend.sum(nvalue * self.kernel_p / norm_p, axis=-1)

        # Sort nodes according to score
        # Then sort after former node ids -> stable = True keeps previous order
        sort1 = tf.argsort(nscore, direction='ASCENDING', stable=False)
        nids_sorted1 = tf.gather(nids, sort1)
        sort2 = tf.argsort(nids_sorted1, direction='ASCENDING', stable=True)  # Must be stable=true here
        sort12 = tf.gather(sort1, sort2)  # index goes from 0 to batch*N, no in batch indexing
        nvalue_sorted = tf.gather(nvalue, sort12, axis=0)
        nscore_sorted = tf.gather(nscore, sort12, axis=0)

        # Make Mask
        nremove = tf.cast(tf.math.round(self.k * tf.cast(nrowlength, dtype=tf.keras.backend.floatx())),
                          dtype=index_dtype)
        nkeep = nrowlength - nremove
        n_remove_keep = ks.backend.flatten(
            tf.concat([ks.backend.expand_dims(nremove, axis=-1), ks.backend.expand_dims(nkeep, axis=-1)], axis=-1))
        mask_remove_keep = ks.backend.flatten(tf.concat(
            [ks.backend.expand_dims(tf.zeros_like(nremove, dtype=tf.bool), axis=-1),
             ks.backend.expand_dims(tf.ones_like(nkeep, tf.bool), axis=-1)], axis=-1))
        mask = tf.repeat(mask_remove_keep, n_remove_keep)

        # Apply Mask to remove lower score nodes
        pooled_n = nvalue_sorted[mask]
        pooled_score = nscore_sorted[mask]
        pooled_id = nids[mask]  # nids should not have changed by final sorting
        pooled_len = nkeep  # shape=(batch,)
        pooled_index = tf.cast(sort12[mask], dtype=index_dtype)  # the index goes from 0 to N*batch
        removed_index = tf.cast(sort12[tf.math.logical_not(mask)],
                                dtype=index_dtype)  # the index goes from 0 to N*batch

        # Pass through gate
        gated_n = pooled_n * ks.backend.expand_dims(tf.keras.activations.sigmoid(pooled_score), axis=-1)

        # Make index map for new nodes towards old index
        index_new_nodes = tf.range(tf.shape(pooled_index)[0], dtype=index_dtype)
        old_shape = tf.cast(ks.backend.expand_dims(tf.shape(nvalue)[0]), dtype=index_dtype)
        map_index = tf.scatter_nd(ks.backend.expand_dims(pooled_index, axis=-1), index_new_nodes, old_shape)

        # Shift index if necessary
        edge_ids = tf.repeat(tf.range(tf.shape(edgelen)[0], dtype=index_dtype), edgelen)

        shiftind = partition_row_indexing(edgeindref, nrowlength, edge_ids,
                                          partition_type_target="row_length", partition_type_index="value_rowids",
                                          from_indexing=self.node_indexing, to_indexing="batch")

        shiftind = tf.cast(shiftind, dtype=index_dtype)  # already shifted by batch offset (sub-graphs)

        # Remove edges that were from filtered nodes via mask
        mask_edge = ks.backend.expand_dims(shiftind, axis=-1) == ks.backend.expand_dims(
            ks.backend.expand_dims(removed_index, axis=0), axis=0)  # this creates large tensor (batch*#edges,2,remove)
        mask_edge = tf.math.logical_not(ks.backend.any(ks.backend.any(mask_edge, axis=-1), axis=-1))
        clean_shiftind = shiftind[mask_edge]
        clean_edge_ids = edge_ids[mask_edge]
        # clean_edge_len = tf.math.segment_sum(tf.ones_like(clean_edge_ids), clean_edge_ids)
        clean_edge_len = tf.scatter_nd(tf.expand_dims(clean_edge_ids, axis=-1), tf.ones_like(clean_edge_ids),
                                       tf.cast(tf.shape(erowlength), dtype=index_dtype))

        # Map edgeindex to new index
        new_edge_index = tf.concat([ks.backend.expand_dims(tf.gather(map_index, clean_shiftind[:, 0]), axis=-1),
                                    ks.backend.expand_dims(tf.gather(map_index, clean_shiftind[:, 1]), axis=-1)],
                                   axis=-1)
        batch_order = tf.argsort(new_edge_index[:, 0], axis=0, direction='ASCENDING', stable=True)
        new_edge_index_sorted = tf.gather(new_edge_index, batch_order, axis=0)

        # Remove the batch offset from edge_indices again for indexing type
        out_indexlist = partition_row_indexing(new_edge_index_sorted,
                                               pooled_len, clean_edge_ids,
                                               partition_type_target="row_length",
                                               partition_type_index="value_rowids",
                                               from_indexing="batch",
                                               to_indexing=self.node_indexing)

        # Correct edge features the same way (remove and reorder)
        edge_feat = edgefeat
        clean_edge_feat = edge_feat[mask_edge]
        clean_edge_feat_sorted = tf.gather(clean_edge_feat, batch_order, axis=0)

        # Make edge feature map for new edge features
        edge_position_old = tf.range(tf.shape(edgefeat)[0], dtype=index_dtype)
        edge_position_new = edge_position_old[mask_edge]
        edge_position_new = tf.gather(edge_position_new, batch_order, axis=0)

        # Collect output tensors
        out_node = gated_n
        out_edge = clean_edge_feat_sorted
        out_edge_index = out_indexlist

        # Change length to partition required
        out_np = pooled_len # row_length
        out_ep = clean_edge_len # row_length

        # Collect reverse pooling info
        # Remove batch offset for old indicies -> but with new length
        out_pool = partition_row_indexing(pooled_index,
                                          nrowlength, pooled_len,
                                          partition_type_target="row_length",
                                          partition_type_index="row_length",
                                          from_indexing="batch",
                                          to_indexing=self.node_indexing)
        out_pool_edge = partition_row_indexing(edge_position_new,
                                               erowlength, clean_edge_ids,
                                               partition_type_target="row_length",
                                               partition_type_index="value_rowids",
                                               from_indexing="batch",
                                               to_indexing=self.node_indexing)

        out = [tf.RaggedTensor.from_row_lengths(out_node, out_np, validate=self.ragged_validate),
               tf.RaggedTensor.from_row_lengths(out_edge, out_ep, validate=self.ragged_validate),
               tf.RaggedTensor.from_row_lengths(out_edge_index, out_ep, validate=self.ragged_validate)]
        out_map = [tf.RaggedTensor.from_row_lengths(out_pool, out_np, validate=self.ragged_validate),
                   tf.RaggedTensor.from_row_lengths(out_pool_edge, out_ep, validate=self.ragged_validate)]

        return out, out_map

    def get_config(self):
        """Update layer config."""
        config = super(PoolingTopK, self).get_config()
        config.update({"k": self.k})
        config.update({
            'kernel_initializer':
                ks.initializers.serialize(self.kernel_initializer),
            'kernel_regularizer':
                ks.regularizers.serialize(self.kernel_regularizer),
            'kernel_constraint':
                ks.constraints.serialize(self.kernel_constraint),
        })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn',name='UnPoolingTopK')
class UnPoolingTopK(GraphBaseLayer):
    """Layer for un-pooling of nodes from PoolingTopK.
    
    The edge index information are not reverted since the tensor before pooling can be reused. 
    Same holds for batch-assignment in number of nodes and edges information.
    """

    def __init__(self, **kwargs):
        """Initialize Layer."""
        super(UnPoolingTopK, self).__init__(**kwargs)

    def build(self, input_shape):
        """Build Layer."""
        super(UnPoolingTopK, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [node, edge, edge_indices, map_node, map_edge, node_pool, edge_pool, edge_indices_pool]

                - node (tf.RaggedTensor): Original node tensor
                - edge (tf.RaggedTensor): Original edge feature tensor
                - edge_indices (tf.RaggedTensor): Original index tensor
                - map_node (tf.RaggedTensor): Index map between original and pooled nodes
                - map_edge (tf.RaggedTensor): Index map between original and pooled edges
                - node_pool (tf.RaggedTensor): Pooled node tensor
                - edge_pool (tf.RaggedTensor): Pooled edge feature tensor
                - edge_indices (tf.RaggedTensor): Pooled index tensor
        
        Returns:
            List: [nodes, edges, edge_indices]
            
                - nodes (tf.RaggedTensor): Un-pooled node feature tensor
                - edges (tf.RaggedTensor): Un-pooled edge feature list
                - edge_indices (tf.RaggedTensor): Un-pooled edge index
        """
        self.assert_ragged_input_rank(inputs)
        dyn_inputs = inputs
        # We cast to values here
        node_old, nrowlength = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
        edge_old, erowlength = dyn_inputs[1].values, dyn_inputs[1].row_lengths()
        edgeind_old, _ = dyn_inputs[2].values, dyn_inputs[2].row_lengths()
        map_node, _ = dyn_inputs[3].values, dyn_inputs[3].row_lengths()
        map_edge, _ = dyn_inputs[4].values, dyn_inputs[4].row_lengths()
        node_new, pool_node_len = dyn_inputs[5].values, dyn_inputs[5].row_lengths()
        edge_new, pool_edge_id = dyn_inputs[6].values, dyn_inputs[6].value_rowids()
        edgeind_new, _ = dyn_inputs[7].values, dyn_inputs[7].row_lengths()

        # Correct map index for flatten batch offset
        map_node = partition_row_indexing(map_node, nrowlength, pool_node_len,
                                          partition_type_target="row_length", partition_type_index="row_length",
                                          from_indexing=self.node_indexing, to_indexing="batch")
        map_edge = partition_row_indexing(map_edge, erowlength, pool_edge_id,
                                          partition_type_target="row_length",
                                          partition_type_index="value_rowids",
                                          from_indexing=self.node_indexing,
                                          to_indexing="batch")

        index_dtype = map_node.dtype
        node_shape = tf.stack([tf.cast(tf.shape(node_old)[0], dtype=index_dtype),
                               tf.cast(tf.shape(node_new)[1], dtype=index_dtype)])
        out_node = tf.scatter_nd(ks.backend.expand_dims(map_node, axis=-1), node_new, node_shape)

        index_dtype = map_edge.dtype
        edge_shape = tf.stack([tf.cast(tf.shape(edge_old)[0], dtype=index_dtype),
                               tf.cast(tf.shape(edge_new)[1], dtype=index_dtype)])
        out_edge = tf.scatter_nd(ks.backend.expand_dims(map_edge, axis=-1), edge_new, edge_shape)

        outlist = [tf.RaggedTensor.from_row_lengths(out_node, nrowlength, validate=self.ragged_validate),
                   tf.RaggedTensor.from_row_lengths(out_edge, erowlength, validate=self.ragged_validate),
                   tf.RaggedTensor.from_row_lengths(edgeind_old, erowlength, validate=self.ragged_validate)]
        return outlist

    def get_config(self):
        """Update layer config."""
        config = super(UnPoolingTopK, self).get_config()
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='AdjacencyPower')
class AdjacencyPower(GraphBaseLayer):
    """Computes powers of the adjacency matrix. This implementation is a temporary solution.

    Note: Layer casts to dense until sparse matmul is supported. This can be very inefficient.

    Args:
        n (int): Power of the adjacency matrix. Default is 2.
    """

    def __init__(self, n=2, **kwargs):
        """Initialize layer."""
        super(AdjacencyPower, self).__init__(**kwargs)
        self.n = n

    def build(self, input_shape):
        """Build layer."""
        super(AdjacencyPower, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, edges, edge_indices]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - edges (tf.RaggedTensor): Adjacency entries of shape (batch, [M], 1)
                - edge_indices (tf.RaggedTensor): Edge-index list referring to nodes of shape (batch, [M], 2)

        Returns:
            list: [edges, edge_indices]

                - edges (tf.RaggedTensor): Adjacency entries of shape  (batch, [M], 1)
                - edge_indices (tf.RaggedTensor): Flatten index list of shape (batch, [M], 2)
        """
        self.assert_ragged_input_rank(inputs)
        dyn_inputs = inputs

        nod, node_len = dyn_inputs[0].values, dyn_inputs[0].row_lengths()
        edge = dyn_inputs[1].values
        edge_index, edge_len = dyn_inputs[2].values, dyn_inputs[2].row_lengths()

        # batch-wise indexing
        edge_index = partition_row_indexing(edge_index,
                                            node_len, edge_len,
                                            partition_type_target="row_length",
                                            partition_type_index="row_length",
                                            from_indexing=self.node_indexing,
                                            to_indexing="sample")

        ind_batch = tf.cast(tf.expand_dims(tf.repeat(tf.range(tf.shape(edge_len)[0]), edge_len), axis=-1),
                            dtype=edge_index.dtype)
        ind_all = tf.concat([ind_batch, edge_index], axis=-1)
        ind_all = tf.cast(ind_all, dtype=tf.int64)

        max_index = tf.reduce_max(node_len)
        dense_shape = tf.stack([tf.cast(tf.shape(node_len)[0], dtype=max_index.dtype), max_index, max_index])
        adj = tf.zeros(dense_shape, dtype=edge.dtype)
        ind_flat = tf.range(tf.cast(tf.shape(node_len)[0], dtype=max_index.dtype) * max_index * max_index)

        adj = tf.expand_dims(adj, axis=-1)
        adj = tf.tensor_scatter_nd_update(adj, ind_all, edge[:, 0:1])
        adj = tf.squeeze(adj, axis=-1)

        out0 = adj
        out = adj
        for i in range(self.n - 1):
            out = tf.linalg.matmul(out, out0)

        # debug_result = out

        # Make sparse
        mask = out > tf.keras.backend.epsilon()
        mask = tf.reshape(mask, (-1,))
        out = tf.reshape(out, (-1,))

        new_edge = out[mask]
        new_edge = tf.expand_dims(new_edge, axis=-1)
        new_indices = tf.unravel_index(ind_flat[mask], dims=dense_shape)
        new_egde_ids = new_indices[0]
        new_edge_index = tf.concat([tf.expand_dims(new_indices[1], axis=-1), tf.expand_dims(new_indices[2], axis=-1)],
                                   axis=-1)
        new_edge_len = tf.tensor_scatter_nd_add(tf.zeros_like(node_len), tf.expand_dims(new_egde_ids, axis=-1),
                                                tf.ones_like(new_egde_ids))

        # batchwise indexing
        new_edge_index = partition_row_indexing(new_edge_index,
                                                node_len, new_edge_len,
                                                partition_type_target="row_length",
                                                partition_type_index="row_length",
                                                from_indexing="sample",
                                                to_indexing=self.node_indexing)

        outlist = [tf.RaggedTensor.from_row_lengths(new_edge, new_edge_len, validate=self.ragged_validate),
                   tf.RaggedTensor.from_row_lengths(new_edge_index, new_edge_len, validate=self.ragged_validate)]

        return outlist

    def get_config(self):
        """Update layer config."""
        config = super(AdjacencyPower, self).get_config()
        config.update({"n": self.n})
        return config


from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    name="kgcnn",
    version="2.0.1",
    author="Patrick Reiser",
    author_email="patrick.reiser@kit.edu",
    description="General Base Layers for Graph Convolutions with tensorflow.keras",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aimat-lab/gcnn_keras",
    install_requires=['numpy', "scikit-learn", "pandas", "scipy", "requests", "matplotlib", "networkx", "sympy",
                      "requests", "tensorflow-addons", "keras-tuner", "pyyaml", "networkx"],
    extras_require={
        "tf": ["tensorflow>=2.4.0"],
        "tf_gpu": ["tensorflow-gpu>=2.4.0"],
    },
    packages=find_packages(),
    include_package_data=True,
    package_data={"kgcnn": ["*.json", "*.yaml"]},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    keywords=["materials", "science", "machine", "learning", "deep", "graph", "networks", "neural"]
)



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.ops.partition import partition_row_indexing
from kgcnn.ops.ragged import partition_from_ragged_tensor_by_name
from kgcnn.layers.base import GraphBaseLayer


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ChangeTensorType')
class ChangeTensorType(GraphBaseLayer):
    """Layer to change the ragged tensor representation into tensor type information.

    The tensor representation can be tf.RaggedTensor, tf.Tensor or a list of (values, partition).
    The RaggedTensor has shape (batch, None, F). The dense tensor in case of equal sized graphs or zero padded
    graphs has shape (batch, N, F).
    For disjoint representation (values, partition), the embeddings are given by
    a flatten value tensor of shape (batch*None, F) and a partition tensor of either "row_length",
    "row_splits" or "value_rowids" that matches the tf.RaggedTensor partition information, which effectively keeps
    the batch assignment.

    Args:
        partition_type (str): Partition tensor type. Default is "row_length".
        input_tensor_type (str): Input type of the tensors for call(). Default is "RaggedTensor".
        output_tensor_type (str): Input type of the tensors for call(). Default is "RaggedTensor".
    """
    def __init__(self,
                 partition_type="row_length",
                 input_tensor_type="RaggedTensor",
                 output_tensor_type="RaggedTensor",
                 **kwargs):
        """Initialize layer."""
        super(ChangeTensorType, self).__init__(**kwargs)
        self.partition_type = partition_type
        self.input_tensor_type = input_tensor_type
        self.output_tensor_type = output_tensor_type

        if self.input_tensor_type not in ["ragged", "RaggedTensor"]:
            raise ValueError("ERROR:kgcnn: Input must be RaggedTensor for layer", self.name)

    def build(self, input_shape):
        """Build layer."""
        super(ChangeTensorType, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Graph tensor.

        Returns:
            tensor: Changed tensor type.
        """
        self.assert_ragged_input_rank(inputs)

        if self.output_tensor_type in ["Tensor", "tensor", "padded", "masked"]:
            return inputs.to_tensor()
        elif self.output_tensor_type in ["ragged", "RaggedTensor"]:
            return inputs
        elif self.output_tensor_type in ["disjoint", "row_partition", "nested", "values_partition"]:
            return partition_from_ragged_tensor_by_name(inputs, self.partition_type)
        else:
            raise NotImplementedError("ERROR:kgcnn: Unsupported output_tensor_type %s" % self.output_tensor_type)

    def get_config(self):
        """Update layer config."""
        config = super(ChangeTensorType, self).get_config()
        config.update({"partition_type": self.partition_type,
                       "input_tensor_type": self.input_tensor_type,
                       "output_tensor_type": self.output_tensor_type
                       })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='ChangeIndexing')
class ChangeIndexing(GraphBaseLayer):
    """Shift the index for index-tensors to assign e.g. nodes within the batch or for each of the sample
    or vice-versa.
    This can be interpreted as a per-graph assignment of batched graph representations versus a global disjoint graph
    representation with unconnected sub-graphs.
    
    .. note::
        Flatten operation changes index tensor as [[0, 1, 2], [0, 1], [0, 1]] to [0, 1, 2, 0, 1, 0, 1] with
        requires a subsequent index-shift of [0, 1, 2, 1, 1, 0, 1] to [0, 1, 2, 3+0, 3+1, 5+0, 5+1].
        This is equivalent to a single graph with disconnected sub-graphs.
    
    Args:
        to_indexing (str): The index refer to the overall 'batch' or to single 'sample'.
            The disjoint representation assigns nodes within the 'batch'.
            It changes "sample" to "batch" or "batch" to "sample." Default is 'batch'.
        from_indexing (str): Index convention that has been set for the input. Default is 'sample'..
        ragged_validate (bool): Whether to validate ragged tensor. Default is False.
    """

    def __init__(self,
                 to_indexing='batch',
                 from_indexing='sample',
                 ragged_validate=False,
                 **kwargs):
        """Initialize layer."""
        super(ChangeIndexing, self).__init__(**kwargs)
        self.to_indexing = to_indexing
        self.from_indexing = from_indexing
        self.ragged_validate = ragged_validate

        if self.from_indexing != self.node_indexing:
            print("WARNING: Graph layer's node_indexing does not agree with from_indexing", self.node_indexing,
                  "vs.", self.to_indexing)

    def build(self, input_shape):
        """Build layer."""
        super(ChangeIndexing, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (list): [nodes, tensor_index]

                - nodes (tf.RaggedTensor): Node embeddings of shape (batch, [N], F)
                - tensor_index (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [N], 2).
            
        Returns:
            tf.RaggedTensor: Batch-wise edge indices of shape (batch, [N], 2).
        """
        self.assert_ragged_input_rank(inputs)
        nod, edge_index = inputs
        indexlist = partition_row_indexing(edge_index.values,
                                           nod.row_splits,
                                           edge_index.value_rowids(),
                                           partition_type_target="row_splits",
                                           partition_type_index="value_rowids",
                                           from_indexing=self.from_indexing,
                                           to_indexing=self.to_indexing)

        out = tf.RaggedTensor.from_row_splits(indexlist, edge_index.row_splits, validate=self.ragged_validate)
        return out

    def get_config(self):
        """Update layer config."""
        config = super(ChangeIndexing, self).get_config()
        config.update({"to_indexing": self.to_indexing,
                       "from_indexing": self.from_indexing,
                       "ragged_validate": self.ragged_validate})
        return config



import unittest
import sympy as sym
import numpy as np

from scipy.optimize import brentq
from scipy import special as sp

import tensorflow as tf
from kgcnn.utils.adj import get_angle_indices
from kgcnn.layers.geom import NodeDistanceEuclidean, EdgeAngle, NodePosition, EdgeAngle2
from kgcnn.layers.conv.dimenet_conv import SphericalBasisLayer
from kgcnn.layers.geom import BesselBasisLayer
from kgcnn.layers.modules import LazySubtract


class TestSphericalBasisLayer(unittest.TestCase):

    def test_result_original(self):

        ei = np.array([[0, 1], [0, 2], [0, 3], [0, 4], [1, 0], [1, 2], [1, 3], [1, 4], [2, 0], [2, 1], [2, 3], [2, 4],
                               [3, 0], [3, 1], [3, 2], [3, 4], [4, 0], [4, 1], [4, 2], [4, 3]])
        ei1 = np.array([[0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7],
                        [0, 8], [0, 9], [0, 10], [1, 0], [1, 2], [1, 3], [1, 4], [1, 5],
                        [1, 6], [1, 7], [1, 8], [1, 9], [1, 10], [2, 0], [2, 1], [2, 3],
                        [2, 4], [2, 5], [2, 6], [2, 7], [2, 8], [2, 9], [2, 10], [3, 0],
                        [3, 1], [3, 2], [3, 4], [3, 5], [3, 6], [3, 7], [3, 8], [3, 9],
                        [3, 10], [4, 0], [4, 1], [4, 2], [4, 3], [4, 5], [4, 6], [4, 7], [4, 8],
                        [4, 9], [4, 10], [5, 0], [5, 1], [5, 2], [5, 3], [5, 4], [5, 6], [5, 7],
                        [5, 8], [5, 10], [6, 0], [6, 1], [6, 2], [6, 3], [6, 4], [6, 5], [6, 7],
                        [6, 8], [6, 9], [6, 10], [7, 0], [7, 1], [7, 2], [7, 3], [7, 4], [7, 5],
                        [7, 6], [7, 8], [7, 9], [7, 10], [8, 0], [8, 1], [8, 2], [8, 3], [8, 4],
                        [8, 5], [8, 6], [8, 7], [8, 9], [8, 10], [9, 0], [9, 1], [9, 2],
                        [9, 3], [9, 4], [9, 6], [9, 7], [9, 8], [9, 10], [10, 0], [10, 1], [10, 2],
                        [10, 3], [10, 4], [10, 5], [10, 6], [10, 7], [10, 8], [10, 9]])
        x1 = np.array([[-0.03113825,  1.54081582,  0.03192126],
               [ 0.01215347,  0.01092235, -0.01603259],
               [ 0.72169129, -0.52583353, -1.2623057 ],
               [ 0.97955987,  1.96459116,  0.03098367],
               [-0.55840223,  1.94831192, -0.83816075],
               [-0.54252252,  1.90153531,  0.93005671],
               [ 0.51522791, -0.36840234,  0.88231134],
               [-1.01070641, -0.38456999,  0.02051783],
               [ 1.7585121 , -0.17376585, -1.30871516],
               [ 0.74087192, -1.62024959, -1.27516511],
               [ 0.22023351, -0.19051179, -2.1772902 ]])
        x= np.array([[-1.26981359e-02,  1.08580416e+00,  8.00099580e-03],
               [ 2.15041600e-03, -6.03131760e-03,  1.97612040e-03],
               [ 1.01173084e+00,  1.46375116e+00,  2.76574800e-04],
               [-5.40815069e-01,  1.44752661e+00, -8.76643715e-01],
               [-5.23813634e-01,  1.43793264e+00,  9.06397294e-01]])

        ei, _, a = get_angle_indices(ei)
        ei1, _, a1 = get_angle_indices(ei1)

        rag_x = tf.RaggedTensor.from_row_lengths(np.concatenate([x,x1]), np.array([len(x),len(x1)]))
        rag_ei = tf.RaggedTensor.from_row_lengths(np.concatenate([ei,ei1]), np.array([len(ei),len(ei1)]))
        rag_a = tf.RaggedTensor.from_row_lengths(np.concatenate([a,a1]), np.array([len(a),len(a1)]))


        class Envelope(tf.keras.layers.Layer):
            """
            Envelope function that ensures a smooth cutoff
            """
            def __init__(self, exponent, name='envelope', **kwargs):
                super().__init__(name=name, **kwargs)
                self.exponent = exponent

                self.p = exponent + 1
                self.a = -(self.p + 1) * (self.p + 2) / 2
                self.b = self.p * (self.p + 2)
                self.c = -self.p * (self.p + 1) / 2

            def call(self, inputs):

                # Envelope function divided by r
                env_val = 1 / inputs + self.a * inputs**(self.p - 1) + self.b * inputs**self.p + self.c * inputs**(self.p + 1)

                return tf.where(inputs < 1, env_val, tf.zeros_like(inputs))


        def Jn(r, n):
            """
            numerical spherical bessel functions of order n
            """
            return np.sqrt(np.pi/(2*r)) * sp.jv(n+0.5, r)


        def Jn_zeros(n, k):
            """
            Compute the first k zeros of the spherical bessel functions up to order n (excluded)
            """
            zerosj = np.zeros((n, k), dtype="float32")
            zerosj[0] = np.arange(1, k + 1) * np.pi
            points = np.arange(1, k + n) * np.pi
            racines = np.zeros(k + n - 1, dtype="float32")
            for i in range(1, n):
                for j in range(k + n - 1 - i):
                    foo = brentq(Jn, points[j], points[j + 1], (i,))
                    racines[j] = foo
                points = racines
                zerosj[i][:k] = racines[:k]

            return zerosj


        def spherical_bessel_formulas(n):
            """
            Computes the sympy formulas for the spherical bessel functions up to order n (excluded)
            """
            x = sym.symbols('x')

            f = [sym.sin(x)/x]
            a = sym.sin(x)/x
            for i in range(1, n):
                b = sym.diff(a, x)/x
                f += [sym.simplify(b*(-x)**i)]
                a = sym.simplify(b)
            return f

        def bessel_basis(n, k):
            """
            Compute the sympy formulas for the normalized and rescaled spherical bessel functions up to
            order n (excluded) and maximum frequency k (excluded).
            """

            zeros = Jn_zeros(n, k)
            normalizer = []
            for order in range(n):
                normalizer_tmp = []
                for i in range(k):
                    normalizer_tmp += [0.5*Jn(zeros[order, i], order+1)**2]
                normalizer_tmp = 1/np.array(normalizer_tmp)**0.5
                normalizer += [normalizer_tmp]

            f = spherical_bessel_formulas(n)
            x = sym.symbols('x')
            bess_basis = []
            for order in range(n):
                bess_basis_tmp = []
                for i in range(k):
                    bess_basis_tmp += [sym.simplify(normalizer[order][i]*f[order].subs(x, zeros[order, i]*x))]
                bess_basis += [bess_basis_tmp]
            return bess_basis


        def sph_harm_prefactor(l, m):
            """
            Computes the constant pre-factor for the spherical harmonic of degree l and order m
            input:
            l: int, l>=0
            m: int, -l<=m<=l
            """
            return ((2*l+1) * np.math.factorial(l-abs(m)) / (4*np.pi*np.math.factorial(l+abs(m))))**0.5


        def associated_legendre_polynomials(l, zero_m_only=True):
            """
            Computes sympy formulas of the associated legendre polynomials up to order l (excluded).
            """
            z = sym.symbols('z')
            P_l_m = [[0]*(j+1) for j in range(l)]

            P_l_m[0][0] = 1
            if l > 0:
                P_l_m[1][0] = z

                for j in range(2, l):
                    P_l_m[j][0] = sym.simplify(
                        ((2*j-1)*z*P_l_m[j-1][0] - (j-1)*P_l_m[j-2][0])/j)
                if not zero_m_only:
                    for i in range(1, l):
                        P_l_m[i][i] = sym.simplify((1-2*i)*P_l_m[i-1][i-1])
                        if i + 1 < l:
                            P_l_m[i+1][i] = sym.simplify((2*i+1)*z*P_l_m[i][i])
                        for j in range(i + 2, l):
                            P_l_m[j][i] = sym.simplify(
                                ((2*j-1) * z * P_l_m[j-1][i] - (i+j-1) * P_l_m[j-2][i]) / (j - i))

            return P_l_m


        def real_sph_harm(l, zero_m_only=True, spherical_coordinates=True):
            """
            Computes formula strings of the the real part of the spherical harmonics up to order l (excluded).
            Variables are either cartesian coordinates x,y,z on the unit sphere or spherical coordinates phi and theta.
            """
            if not zero_m_only:
                S_m = [0]
                C_m = [1]
                for i in range(1, l):
                    x = sym.symbols('x')
                    y = sym.symbols('y')
                    S_m += [x*S_m[i-1] + y*C_m[i-1]]
                    C_m += [x*C_m[i-1] - y*S_m[i-1]]

            P_l_m = associated_legendre_polynomials(l, zero_m_only)
            if spherical_coordinates:
                theta = sym.symbols('theta')
                z = sym.symbols('z')
                for i in range(len(P_l_m)):
                    for j in range(len(P_l_m[i])):
                        if type(P_l_m[i][j]) != int:
                            P_l_m[i][j] = P_l_m[i][j].subs(z, sym.cos(theta))
                if not zero_m_only:
                    phi = sym.symbols('phi')
                    for i in range(len(S_m)):
                        S_m[i] = S_m[i].subs(x, sym.sin(
                            theta)*sym.cos(phi)).subs(y, sym.sin(theta)*sym.sin(phi))
                    for i in range(len(C_m)):
                        C_m[i] = C_m[i].subs(x, sym.sin(
                            theta)*sym.cos(phi)).subs(y, sym.sin(theta)*sym.sin(phi))

            Y_func_l_m = [['0']*(2*j + 1) for j in range(l)]
            for i in range(l):
                Y_func_l_m[i][0] = sym.simplify(sph_harm_prefactor(i, 0) * P_l_m[i][0])

            if not zero_m_only:
                for i in range(1, l):
                    for j in range(1, i + 1):
                        Y_func_l_m[i][j] = sym.simplify(
                            2**0.5 * sph_harm_prefactor(i, j) * C_m[j] * P_l_m[i][j])
                for i in range(1, l):
                    for j in range(1, i + 1):
                        Y_func_l_m[i][-j] = sym.simplify(
                            2**0.5 * sph_harm_prefactor(i, -j) * S_m[j] * P_l_m[i][j])

            return Y_func_l_m


        class SphericalBasisLayerOrginal(tf.keras.layers.Layer):
            def __init__(self, num_spherical, num_radial, cutoff, envelope_exponent=5,
                         name='spherical_basis', **kwargs):
                super().__init__(name=name, **kwargs)

                assert num_radial <= 64
                self.num_radial = num_radial
                self.num_spherical = num_spherical

                self.inv_cutoff = tf.constant(1 / cutoff, dtype=tf.float32)
                self.envelope = Envelope(envelope_exponent)

                # retrieve formulas
                self.bessel_formulas = bessel_basis(num_spherical, num_radial)
                self.sph_harm_formulas = real_sph_harm(num_spherical)
                self.sph_funcs = []
                self.bessel_funcs = []

                # convert to tensorflow functions
                x = sym.symbols('x')
                theta = sym.symbols('theta')
                for i in range(num_spherical):
                    if i == 0:
                        first_sph = sym.lambdify([theta], self.sph_harm_formulas[i][0], 'tensorflow')(0)
                        self.sph_funcs.append(lambda tensor: tf.zeros_like(tensor) + first_sph)
                    else:
                        self.sph_funcs.append(sym.lambdify([theta], self.sph_harm_formulas[i][0], 'tensorflow'))
                    for j in range(num_radial):
                        self.bessel_funcs.append(sym.lambdify([x], self.bessel_formulas[i][j], 'tensorflow'))

            def call(self, inputs):
                d, Angles, id_expand_kj = inputs

                d_scaled = d * self.inv_cutoff
                rbf = [f(d_scaled) for f in self.bessel_funcs]
                rbf = tf.stack(rbf, axis=1)

                d_cutoff = self.envelope(d_scaled)
                rbf_env = d_cutoff[:, None] * rbf
                rbf_env = tf.gather(rbf_env, id_expand_kj)

                cbf = [f(Angles) for f in self.sph_funcs]
                cbf = tf.stack(cbf, axis=1)
                cbf = tf.repeat(cbf, self.num_radial, axis=1)

                return cbf*rbf_env

        a, b = NodePosition()([rag_x, rag_ei])
        dist = NodeDistanceEuclidean()([a, b])
        vec = LazySubtract()([a, b])
        angs = EdgeAngle()([vec, rag_a])
        bessel = SphericalBasisLayer(10,10,5.0)([dist, angs, rag_a])

        test = np.array(SphericalBasisLayerOrginal(10,10,5)([dist[0,:,0], angs[0,:,0], rag_a[0,:,1]]))
        test1 = np.array(SphericalBasisLayerOrginal(10,10,5)([dist[1,:,0], angs[1,:,0], rag_a[1,:,1]]))
        pred = np.array(bessel[0])
        pred1 = np.array(bessel[1])

        valid = np.abs(test - pred)
        valid1 = np.abs(test1 - pred1)
        # print(np.max(valid))
        # print(np.max(valid1))

        # For small argument and large order there are some rof errors
        # So we only have < 0.1 here
        self.assertTrue(np.max(valid) < 0.05)
        self.assertTrue(np.max(valid1) < 0.05)


class TestBesselBasisLayer(unittest.TestCase):

    def test_result_original(self):

        ei = np.array([[0, 1], [0, 2], [0, 3], [0, 4], [1, 0], [1, 2], [1, 3], [1, 4], [2, 0], [2, 1], [2, 3], [2, 4],
                       [3, 0], [3, 1], [3, 2], [3, 4], [4, 0], [4, 1], [4, 2], [4, 3]])
        ei1 = np.array([[0, 1], [0, 2], [0, 3], [0, 4], [0, 5], [0, 6], [0, 7],
                        [0, 8], [0, 9], [0, 10], [1, 0], [1, 2], [1, 3], [1, 4], [1, 5],
                        [1, 6], [1, 7], [1, 8], [1, 9], [1, 10], [2, 0], [2, 1], [2, 3],
                        [2, 4], [2, 5], [2, 6], [2, 7], [2, 8], [2, 9], [2, 10], [3, 0],
                        [3, 1], [3, 2], [3, 4], [3, 5], [3, 6], [3, 7], [3, 8], [3, 9],
                        [3, 10], [4, 0], [4, 1], [4, 2], [4, 3], [4, 5], [4, 6], [4, 7], [4, 8],
                        [4, 9], [4, 10], [5, 0], [5, 1], [5, 2], [5, 3], [5, 4], [5, 6], [5, 7],
                        [5, 8], [5, 10], [6, 0], [6, 1], [6, 2], [6, 3], [6, 4], [6, 5], [6, 7],
                        [6, 8], [6, 9], [6, 10], [7, 0], [7, 1], [7, 2], [7, 3], [7, 4], [7, 5],
                        [7, 6], [7, 8], [7, 9], [7, 10], [8, 0], [8, 1], [8, 2], [8, 3], [8, 4],
                        [8, 5], [8, 6], [8, 7], [8, 9], [8, 10], [9, 0], [9, 1], [9, 2],
                        [9, 3], [9, 4], [9, 6], [9, 7], [9, 8], [9, 10], [10, 0], [10, 1], [10, 2],
                        [10, 3], [10, 4], [10, 5], [10, 6], [10, 7], [10, 8], [10, 9]])
        x1 = np.array([[-0.03113825,  1.54081582,  0.03192126],
               [ 0.01215347,  0.01092235, -0.01603259],
               [ 0.72169129, -0.52583353, -1.2623057 ],
               [ 0.97955987,  1.96459116,  0.03098367],
               [-0.55840223,  1.94831192, -0.83816075],
               [-0.54252252,  1.90153531,  0.93005671],
               [ 0.51522791, -0.36840234,  0.88231134],
               [-1.01070641, -0.38456999,  0.02051783],
               [ 1.7585121 , -0.17376585, -1.30871516],
               [ 0.74087192, -1.62024959, -1.27516511],
               [ 0.22023351, -0.19051179, -2.1772902 ]])
        x= np.array([[-1.26981359e-02,  1.08580416e+00,  8.00099580e-03],
               [ 2.15041600e-03, -6.03131760e-03,  1.97612040e-03],
               [ 1.01173084e+00,  1.46375116e+00,  2.76574800e-04],
               [-5.40815069e-01,  1.44752661e+00, -8.76643715e-01],
               [-5.23813634e-01,  1.43793264e+00,  9.06397294e-01]])


        rag_x = tf.RaggedTensor.from_row_lengths(np.concatenate([x,x1]), np.array([len(x),len(x1)]))
        rag_ei = tf.RaggedTensor.from_row_lengths(np.concatenate([ei,ei1]), np.array([len(ei),len(ei1)]))
        a, b = NodePosition()([rag_x, rag_ei])
        dist = NodeDistanceEuclidean()([a, b])
        bessel = BesselBasisLayer(10,5.0)(dist)

        class Envelope(tf.keras.layers.Layer):
            """
            Envelope function that ensures a smooth cutoff
            """
            def __init__(self, exponent, name='envelope', **kwargs):
                super().__init__(name=name, **kwargs)
                self.exponent = exponent

                self.p = exponent + 1
                self.a = -(self.p + 1) * (self.p + 2) / 2
                self.b = self.p * (self.p + 2)
                self.c = -self.p * (self.p + 1) / 2

            def call(self, inputs):

                # Envelope function divided by r
                env_val = 1 / inputs + self.a * inputs**(self.p - 1) + self.b * inputs**self.p + self.c * inputs**(self.p + 1)

                return tf.where(inputs < 1, env_val, tf.zeros_like(inputs))


        class BesselBasisLayerOriginal(tf.keras.layers.Layer):
            def __init__(self, num_radial, cutoff, envelope_exponent=5,
                         name='bessel_basis', **kwargs):
                super().__init__(name=name, **kwargs)
                self.num_radial = num_radial
                self.inv_cutoff = tf.constant(1 / cutoff, dtype=tf.float32)
                self.envelope = Envelope(envelope_exponent)

                # Initialize frequencies at canonical positions
                def freq_init(shape, dtype):
                    return tf.constant(np.pi * np.arange(1, shape + 1, dtype=np.float32), dtype=dtype)
                self.frequencies = self.add_weight(name="frequencies", shape=self.num_radial,
                                                   dtype=tf.float32, initializer=freq_init, trainable=True)

            def call(self, inputs):
                d_scaled = inputs * self.inv_cutoff

                # Necessary for proper broadcasting behaviour
                d_scaled = tf.expand_dims(d_scaled, -1)

                d_cutoff = self.envelope(d_scaled)
                return d_cutoff * tf.sin(self.frequencies * d_scaled)

        test0 = BesselBasisLayerOriginal(10,5.0)(dist[0,:,0])
        test1 = BesselBasisLayerOriginal(10,5.0)(dist[1,:,0])

        self.assertTrue(np.max(np.abs(test0 - bessel[0])) < 1e-5)
        self.assertTrue(np.max(np.abs(test1 - bessel[1])) < 1e-5)


if __name__ == '__main__':
    unittest.main()



import tensorflow.keras as ks

from kgcnn.layers.conv.megnet_conv import MEGnetBlock
from kgcnn.layers.geom import NodeDistanceEuclidean, GaussBasisLayer, NodePosition
from kgcnn.layers.modules import DenseEmbedding, LazyAdd, DropoutEmbedding, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingGlobalEdges, PoolingNodes
from kgcnn.layers.pool.set2set import PoolingSet2Set
from kgcnn.utils.models import update_model_kwargs

# from kgcnn.layers.casting import ChangeTensorType, ChangeIndexing

# Graph Networks as a Universal Machine Learning Framework for Molecules and Crystals
# by Chi Chen, Weike Ye, Yunxing Zuo, Chen Zheng, and Shyue Ping Ong*
# https://github.com/materialsvirtuallab/megnet


model_default = {'name': "Megnet",
                 'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 3), 'name': "node_coordinates", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True},
                            {'shape': [], 'name': "graph_attributes", 'dtype': 'float32', 'ragged': False}],
                 'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                     "graph": {"input_dim": 100, "output_dim": 64}},
                 "make_distance": True, "expand_distance": True,
                 'gauss_args': {"bins": 20, "distance": 4, "offset": 0.0, "sigma": 0.4},
                 'meg_block_args': {'node_embed': [64, 32, 32], 'edge_embed': [64, 32, 32],
                                    'env_embed': [64, 32, 32], 'activation': 'kgcnn>softplus2'},
                 'set2set_args': {'channels': 16, 'T': 3, "pooling_method": "sum", "init_qstar": "0"},
                 'node_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
                 'edge_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
                 'state_ff_args': {"units": [64, 32], "activation": "kgcnn>softplus2"},
                 'nblocks': 3, 'has_ff': True, 'dropout': None, 'use_set2set': True,
                 'verbose': 10,
                 'output_embedding': 'graph',
                 'output_mlp': {"use_bias": [True, True, True], "units": [32, 16, 1],
                                "activation": ['kgcnn>softplus2', 'kgcnn>softplus2', 'linear']}
                 }


@update_model_kwargs(model_default)
def make_model(inputs=None,
               input_embedding=None,
               expand_distance=None,
               make_distance=None,
               gauss_args=None,
               meg_block_args=None,
               set2set_args=None,
               node_ff_args=None,
               edge_ff_args=None,
               state_ff_args=None,
               use_set2set=None,
               nblocks=None,
               has_ff=None,
               dropout=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make MegNet graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        make_distance (bool): Whether input is distance or coordinates at in place of edges.
        expand_distance (bool): If the edge input are actual edges or node coordinates instead that are expanded to
            form edges with a gauss distance basis given edge indices indices. Expansion uses `gauss_args`.
        gauss_args (dict): Dictionary of layer arguments unpacked in `GaussBasisLayer` layer.
        meg_block_args (dict): Dictionary of layer arguments unpacked in `MEGnetBlock` layer.
        set2set_args (dict): Dictionary of layer arguments unpacked in `PoolingSet2Set` layer.
        node_ff_args (dict): Dictionary of layer arguments unpacked in `MLP` feed-forward layer.
        edge_ff_args (dict): Dictionary of layer arguments unpacked in `MLP` feed-forward layer.
        state_ff_args (dict): Dictionary of layer arguments unpacked in `MLP` feed-forward layer.
        use_set2set (bool): Whether to use `PoolingSet2Set` layer.
        nblocks (int): Number of graph embedding blocks or depth of the network.
        has_ff (bool): Use feed-forward MLP in each block.
        dropout (int): Dropout to use. Default is None.
        name (str): Name of the model.
        verbose (int): Verbosity level of print.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    env_input = ks.Input(**inputs[3])

    # embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    uenv = OptionalInputEmbedding(**input_embedding['graph'],
                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
    edi = edge_index_input

    # Edge distance as Gauss-Basis
    if make_distance:
        x = xyz_input
        pos1, pos2 = NodePosition()([x, edi])
        ed = NodeDistanceEuclidean()([pos1, pos2])
    else:
        ed = xyz_input

    if expand_distance:
        ed = GaussBasisLayer(**gauss_args)(ed)

    # Model
    vp = n
    ep = ed
    up = uenv
    vp = GraphMLP(**node_ff_args)(vp)
    ep = GraphMLP(**edge_ff_args)(ep)
    up = MLP(**state_ff_args)(up)
    vp2 = vp
    ep2 = ep
    up2 = up
    for i in range(0, nblocks):
        if has_ff and i > 0:
            vp2 = GraphMLP(**node_ff_args)(vp)
            ep2 = GraphMLP(**edge_ff_args)(ep)
            up2 = MLP(**state_ff_args)(up)

        # MEGnetBlock
        vp2, ep2, up2 = MEGnetBlock(**meg_block_args)(
            [vp2, ep2, edi, up2])

        # skip connection
        if dropout is not None:
            vp2 = DropoutEmbedding(dropout, name='dropout_atom_%d' % i)(vp2)
            ep2 = DropoutEmbedding(dropout, name='dropout_bond_%d' % i)(ep2)
            up2 = ks.layers.Dropout(dropout, name='dropout_state_%d' % i)(up2)

        vp = LazyAdd()([vp2, vp])
        ep = LazyAdd()([ep2, ep])
        up = ks.layers.Add()([up2, up])

    if use_set2set:
        vp = DenseEmbedding(set2set_args["channels"], activation='linear')(vp)  # to match units
        ep = DenseEmbedding(set2set_args["channels"], activation='linear')(ep)  # to match units
        vp = PoolingSet2Set(**set2set_args)(vp)
        ep = PoolingSet2Set(**set2set_args)(ep)
    else:
        vp = PoolingNodes()(vp)
        ep = PoolingGlobalEdges()(ep)

    ep = ks.layers.Flatten()(ep)
    vp = ks.layers.Flatten()(vp)
    final_vec = ks.layers.Concatenate(axis=-1)([vp, ep, up])

    if dropout is not None:
        final_vec = ks.layers.Dropout(dropout, name='dropout_final')(final_vec)

    # Only graph embedding for Megnet
    if output_embedding != "graph":
        raise ValueError("Unsupported graph embedding for mode `Megnet`.")

    main_output = MLP(**output_mlp)(final_vec)
    model = ks.models.Model(inputs=[node_input, xyz_input, edge_index_input, env_input], outputs=main_output)
    return model



import tensorflow as tf
import tensorflow.keras as ks
import numpy as np


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='shifted_softplus')
def shifted_softplus(x):
    r"""Shifted softplus activation function.
    
    Args:
        x (tf.Tensor): Single values to apply activation with tf.keras functions.
    
    Returns:
        tf.Tensor: Output tensor computed as :math:`\log(e^{x}+1) - \log(2)`.
    """
    return ks.activations.softplus(x) - ks.backend.log(2.0)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='softplus2')
def softplus2(x):
    r"""Softplus function that is :math:`0` at :math:`x=0`, the implementation aims at avoiding overflow
    :math:`\log(e^{x}+1) - \log(2)`.
    
    Args:
        x (tf.Tensor): Single values to apply activation with tf.keras functions.
    
    Returns:
         tf.Tensor: Output tensor computed as :math:`\log(e^{x}+1) - \log(2)`.
    """
    return ks.backend.relu(x) + ks.backend.log(0.5 * ks.backend.exp(-ks.backend.abs(x)) + 0.5)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='leaky_softplus')
class leaky_softplus(tf.keras.layers.Layer):
    r"""Leaky softplus activation function similar to :obj:`tf.nn.leaky_relu` but smooth. """

    def __init__(self, alpha=0.05, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            alpha (float, optional): Leak parameter alpha. Default is 0.05.
        """
        super(leaky_softplus, self).__init__(**kwargs)
        # self.alpha = self.add_weight(shape=None, dtype=self.dtype, trainable=trainable)
        # self.set_weights([np.array(alpha)])
        self.alpha = float(alpha)

    def call(self, inputs, **kwargs):
        """Compute leaky_softplus activation from inputs."""
        x = inputs
        return ks.activations.softplus(x) * (1 - self.alpha) + self.alpha * x

    def get_config(self):
        config = super(leaky_softplus, self).get_config()
        config.update({"alpha": self.alpha})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='leaky_relu')
class leaky_relu(tf.keras.layers.Layer):
    r"""Leaky RELU function. Equivalent to :obj:`tf.nn.leaky_relu(x,alpha)`."""

    def __init__(self, alpha: float = 0.05, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            alpha (float, optional): Leak parameter alpha. Default is 0.05.
        """
        super(leaky_relu, self).__init__(**kwargs)
        self.alpha = float(alpha)

    def call(self, inputs, **kwargs):
        """Compute leaky_relu activation from inputs."""
        x = inputs
        return tf.nn.leaky_relu(x, alpha=self.alpha)
        # return tf.nn.relu(x) - tf.nn.relu(-x)*self.alpha

    def get_config(self):
        config = super(leaky_relu, self).get_config()
        config.update({"alpha": self.alpha})
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='swish')
class swish(tf.keras.layers.Layer):
    r"""Swish activation function. Computes :math:`x \; \text{sig}(\beta x)`,
    with :math:`\text{sig}(x) = 1/(1+e^{-x})`."""

    def __init__(self, beta: float = 1.0, **kwargs):
        """Initialize with optionally learnable parameter.

        Args:
            beta (float, optional): Parameter beta in sigmoid. Default is 1.0.
        """
        super(swish, self).__init__(**kwargs)
        # self.beta = self.add_weight(shape=None, dtype=self.dtype, trainable=trainable)
        # self.set_weights([np.array(beta)])
        self.beta = float(beta)

    def call(self, inputs, **kwargs):
        """Compute swish activation from inputs."""
        x = inputs
        return x * tf.sigmoid(self.beta*x)

    def get_config(self):
        config = super(swish, self).get_config()
        config.update({"beta": self.beta})
        return config



import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.gather import GatherState, GatherNodesIngoing, GatherNodesOutgoing
from kgcnn.layers.modules import LazyConcatenate, DenseEmbedding, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingLocalEdges, PoolingNodes
from kgcnn.layers.pool.set2set import PoolingSet2Set
from kgcnn.utils.models import update_model_kwargs

# 'Interaction Networks for Learning about Objects,Relations and Physics'
# by Peter W. Battaglia, Razvan Pascanu, Matthew Lai, Danilo Rezende, Koray Kavukcuoglu
# http://papers.nips.cc/paper/6417-interaction-networks-for-learning-about-objects-relations-and-physics
# https://github.com/higgsfield/interaction_network_pytorch

hyper_model_default = {'name': "INorp",
                       'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                                  {'shape': (None,), 'name': "edge_attributes", 'dtype': 'float32', 'ragged': True},
                                  {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True},
                                  {'shape': [], 'name': "graph_attributes", 'dtype': 'float32', 'ragged': False}],
                       'input_embedding': {"node": {"input_dim": 95, "output_dim": 64},
                                           "edge": {"input_dim": 5, "output_dim": 64},
                                           "graph": {"input_dim": 100, "output_dim": 64}},
                       'set2set_args': {"channels": 32, "T": 3, "pooling_method": "mean",
                                        "init_qstar": "mean"},
                       'node_mlp_args': {"units": [100, 50], "use_bias": True, "activation": ['relu', "linear"]},
                       'edge_mlp_args': {"units": [100, 100, 100, 100, 50],
                                         "activation": ['relu', 'relu', 'relu', 'relu', "linear"]},
                       'pooling_args': {'pooling_method': "segment_mean"},
                       'depth': 3, 'use_set2set': False, 'verbose': 10,
                       'gather_args': {},
                       'output_embedding': 'graph',
                       'output_mlp': {"use_bias": [True, True, False], "units": [25, 10, 1],
                                      "activation": ['relu', 'relu', 'sigmoid']}
                       }


@update_model_kwargs(hyper_model_default)
def make_model(inputs=None,
               input_embedding=None,
               depth=None,
               gather_args=None,
               edge_mlp_args=None,
               node_mlp_args=None,
               set2set_args=None,
               pooling_args=None,
               use_set2set=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make INorp graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        gather_args (dict): Dictionary of layer arguments unpacked in `GatherNodes` layer.
        edge_mlp_args (dict): Dictionary of layer arguments unpacked in `MLP` layer for edge updates.
        node_mlp_args (dict): Dictionary of layer arguments unpacked in `MLP` layer for node updates.
        set2set_args (dict): Dictionary of layer arguments unpacked in `PoolingSet2Set` layer.
        pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingLocalEdges`, `PoolingNodes` layer.
        use_set2set (bool): Whether to use `PoolingSet2Set` layer.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    edge_input = ks.layers.Input(**inputs[1])
    edge_index_input = ks.layers.Input(**inputs[2])
    env_input = ks.Input(**inputs[3])

    # embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    ed = OptionalInputEmbedding(**input_embedding['edge'],
                                use_embedding=len(inputs[1]['shape']) < 2)(edge_input)
    uenv = OptionalInputEmbedding(**input_embedding['graph'],
                                  use_embedding=len(inputs[3]['shape']) < 1)(env_input)
    edi = edge_index_input

    # Model
    ev = GatherState(**gather_args)([uenv, n])
    # n-Layer Step
    for i in range(0, depth):
        # upd = GatherNodes()([n,edi])
        eu1 = GatherNodesIngoing(**gather_args)([n, edi])
        eu2 = GatherNodesOutgoing(**gather_args)([n, edi])
        upd = LazyConcatenate(axis=-1)([eu2, eu1])
        eu = LazyConcatenate(axis=-1)([upd, ed])

        eu = GraphMLP(**edge_mlp_args)(eu)
        # Pool message
        nu = PoolingLocalEdges(**pooling_args)(
            [n, eu, edi])  # Summing for each node connection
        # Add environment
        nu = LazyConcatenate(axis=-1)(
            [n, nu, ev])  # LazyConcatenate node features with new edge updates
        n = GraphMLP(**node_mlp_args)(nu)

    # Output embedding choice
    if output_embedding == 'graph':
        if use_set2set:
            # output
            n = DenseEmbedding(set2set_args["channels"], activation="linear")(n)
            out = PoolingSet2Set(**set2set_args)(n)
        else:
            out = PoolingNodes(**pooling_args)(n)
        out = ks.layers.Flatten()(out)
        main_output = MLP(**output_mlp)(out)
    elif output_embedding == 'node':
        out = n
        main_output = GraphMLP(**output_mlp)(out)
        main_output = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(main_output)
        # no ragged for distribution atm
    else:
        raise ValueError("Unsupported graph embedding for mode `INorp`")

    model = ks.models.Model(inputs=[node_input, edge_input, edge_index_input, env_input], outputs=main_output)
    return model



import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.conv.painn_conv import PAiNNUpdate, EquivariantInitialize
from kgcnn.layers.conv.painn_conv import PAiNNconv
from kgcnn.layers.geom import NodeDistanceEuclidean, BesselBasisLayer, EdgeDirectionNormalized, CosCutOffEnvelope, NodePosition
from kgcnn.layers.modules import LazyAdd, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.utils.models import update_model_kwargs

# Equivariant message passing for the prediction of tensorial properties and molecular spectra
# Kristof T. Schuett, Oliver T. Unke and Michael Gastegger
# https://arxiv.org/pdf/2102.03150.pdf

model_default = {"name": "PAiNN",
                 "inputs": [{"shape": (None,), "name": "node_attributes", "dtype": "float32", "ragged": True},
                            {"shape": (None, 3), "name": "node_coordinates", "dtype": "float32", "ragged": True},
                            {"shape": (None, 2), "name": "edge_indices", "dtype": "int64", "ragged": True}],
                 "input_embedding": {"node": {"input_dim": 95, "output_dim": 128}},
                 "bessel_basis": {"num_radial": 20, "cutoff": 5.0, "envelope_exponent": 5},
                 "pooling_args": {"pooling_method": "sum"},
                 "conv_args": {"units": 128, "cutoff": None, "conv_pool": "sum"},
                 "update_args": {"units": 128},
                 "depth": 3,
                 "verbose": 10,
                 "output_embedding": "graph",
                 "output_mlp": {"use_bias": [True, True], "units": [128, 1], "activation": ["swish", "linear"]}
                 }


@update_model_kwargs(model_default)
def make_model(inputs=None,
               input_embedding=None,
               bessel_basis=None,
               depth=None,
               pooling_args=None,
               conv_args=None,
               update_args=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make PAiNN graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        bessel_basis (dict): Dictionary of layer arguments unpacked in final `BesselBasisLayer` layer.
        depth (int): Number of graph embedding units or depth of the network.
        pooling_args (dict): Dictionary of layer arguments unpacked in `PoolingNodes` layer.
        conv_args (dict): Dictionary of layer arguments unpacked in `PAiNNconv` layer.
        update_args (dict): Dictionary of layer arguments unpacked in `PAiNNUpdate` layer.
        verbose (int): Level of verbosity.
        name (str): Name of the model.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    node_input = ks.layers.Input(**inputs[0])
    xyz_input = ks.layers.Input(**inputs[1])
    bond_index_input = ks.layers.Input(**inputs[2])
    z = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)

    if len(inputs) > 3:
        equiv_input = ks.layers.Input(**inputs[3])
    else:
        equiv_input = EquivariantInitialize(dim=3)(z)

    edi = bond_index_input
    x = xyz_input
    v = equiv_input

    pos1, pos2 = NodePosition()([x, edi])
    rij = EdgeDirectionNormalized()([pos1, pos2])
    d = NodeDistanceEuclidean()([pos1, pos2])
    env = CosCutOffEnvelope(conv_args["cutoff"])(d)
    rbf = BesselBasisLayer(**bessel_basis)(d)

    for i in range(depth):
        # Message
        ds, dv = PAiNNconv(**conv_args)([z, v, rbf, env, rij, edi])
        z = LazyAdd()([z, ds])
        v = LazyAdd()([v, dv])
        # Update
        ds, dv = PAiNNUpdate(**update_args)([z, v])
        z = LazyAdd()([z, ds])
        v = LazyAdd()([v, dv])
    n = z
    # Output embedding choice
    if output_embedding == "graph":
        out = PoolingNodes(**pooling_args)(n)
        main_output = MLP(**output_mlp)(out)
    elif output_embedding == "node":
        out = n
        main_output = GraphMLP(**output_mlp)(out)
        main_output = ChangeTensorType(input_tensor_type="ragged", output_tensor_type="tensor")(main_output)
        # no ragged for distribution atm
    else:
        raise ValueError("Unsupported graph embedding for mode `PAiNN`")

    if len(inputs) > 3:
        model = tf.keras.models.Model(inputs=[node_input, xyz_input, bond_index_input, equiv_input],
                                      outputs=main_output)
    else:
        model = tf.keras.models.Model(inputs=[node_input, xyz_input, bond_index_input],
                                      outputs=main_output)
    return model



import tensorflow as tf

from kgcnn.layers.modules import DenseEmbedding, ActivationEmbedding, DropoutEmbedding
from kgcnn.layers.norm import GraphBatchNormalization, GraphLayerNormalization
from kgcnn.layers.base import GraphBaseLayer
import kgcnn.ops.activ


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='MLPBase')
class MLPBase(GraphBaseLayer):
    r"""Multilayer perceptron that consist of N dense keras layers. Supply list in place of arguments for each layer.
    If not list, then the single argument is used for each layer.
    The number of layers is given by units, which should be list.
    Additionally this base class holds arguments for batch-normalization which should be applied between kernel
    and activation. And dropout after the kernel output and before normalization.
    This base class does not initialize any sub-layers or implements :obj:`call()`. Only for managing arguments.

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            If you don't specify anything, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, whether the layer uses a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix.
        bias_initializer: Initializer for the bias vector.
        kernel_regularizer: Regularizer function applied to
            the `kernel` weights matrix.
        bias_regularizer: Regularizer function applied to the bias vector.
        activity_regularizer: Regularizer function applied to
            the output of the layer (its "activation").
        kernel_constraint: Constraint function applied to
            the `kernel` weights matrix.
        bias_constraint: Constraint function applied to the bias vector.
        use_normalization: Whether to use a normalization layer in between.
        normalization_technique: Which keras normalization technique to apply.
            This can be either 'batch', 'layer', 'group' etc.
        axis: Integer, the axis that should be normalized (typically the features
            axis). For instance, after a `Conv2D` layer with
            `data_format="channels_first"`, set `axis=1` in `GraphBatchNormalization`.
        momentum: Momentum for the moving average.
        epsilon: Small float added to variance to avoid dividing by zero.
        center: If True, add offset of `beta` to normalized tensor. If False, `beta`
            is ignored.
        scale: If True, multiply by `gamma`. If False, `gamma` is not used. When the
            next layer is linear (also e.g. `nn.relu`), this can be disabled since the
            scaling will be done by the next layer.
        beta_initializer: Initializer for the beta weight.
        gamma_initializer: Initializer for the gamma weight.
        moving_mean_initializer: Initializer for the moving mean.
        moving_variance_initializer: Initializer for the moving variance.
        beta_regularizer: Optional regularizer for the beta weight.
        gamma_regularizer: Optional regularizer for the gamma weight.
        beta_constraint: Optional constraint for the beta weight.
        gamma_constraint: Optional constraint for the gamma weight.
        use_dropout: Whether to use dropout layers in between.
        rate: Float between 0 and 1. Fraction of the input units to drop.
        noise_shape: 1D integer tensor representing the shape of the
            binary dropout mask that will be multiplied with the input.
            For instance, if your inputs have shape`(batch_size, timesteps, features)` and
            you want the dropout mask to be the same for all timesteps,
            you can use `noise_shape=(batch_size, 1, features)`.
        seed: A Python integer to use as random seed.
    """

    def __init__(self,
                 units,
                 use_bias=True,
                 activation=None,
                 activity_regularizer=None,
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_constraint=None,
                 bias_constraint=None,
                 # Normalization
                 use_normalization=False,
                 normalization_technique="batch",
                 axis=-1, momentum=0.99, epsilon=0.001, center=True, scale=True,
                 beta_initializer='zeros', gamma_initializer='ones',
                 moving_mean_initializer='zeros',
                 moving_variance_initializer='ones', beta_regularizer=None,
                 gamma_regularizer=None, beta_constraint=None, gamma_constraint=None,
                 # Dropout
                 use_dropout=False,
                 rate=None, noise_shape=None, seed=None,
                 **kwargs):
        """Initialize MLP as for dense."""
        super(MLPBase, self).__init__(**kwargs)
        # everything should be defined by units.
        if isinstance(units, int):
            units = [units]
        if not isinstance(units, list):
            raise ValueError("Units must be a list or a single int for `MLP`.")

        self._depth = len(units)

        # Assert matching number of args
        def assert_args_is_list(args):
            if not isinstance(args, (list, tuple)):
                return [args for _ in range(self._depth)]
            return args

        # Dense
        use_bias = assert_args_is_list(use_bias)
        activation = assert_args_is_list(activation)
        kernel_regularizer = assert_args_is_list(kernel_regularizer)
        bias_regularizer = assert_args_is_list(bias_regularizer)
        activity_regularizer = assert_args_is_list(activity_regularizer)
        kernel_initializer = assert_args_is_list(kernel_initializer)
        bias_initializer = assert_args_is_list(bias_initializer)
        kernel_constraint = assert_args_is_list(kernel_constraint)
        bias_constraint = assert_args_is_list(bias_constraint)
        # Normalization
        use_normalization = assert_args_is_list(use_normalization)
        normalization_technique = assert_args_is_list(normalization_technique)
        if not isinstance(axis, list):  # Special case, if axis is supposed to be multiple axis, use tuple here.
            axis = [axis for _ in units]
        momentum = assert_args_is_list(momentum)
        epsilon = assert_args_is_list(epsilon)
        center = assert_args_is_list(center)
        scale = assert_args_is_list(scale)
        beta_initializer = assert_args_is_list(beta_initializer)
        gamma_initializer = assert_args_is_list(gamma_initializer)
        moving_mean_initializer = assert_args_is_list(moving_mean_initializer)
        moving_variance_initializer = assert_args_is_list(moving_variance_initializer)
        beta_regularizer = assert_args_is_list(beta_regularizer)
        gamma_regularizer = assert_args_is_list(gamma_regularizer)
        beta_constraint = assert_args_is_list(beta_constraint)
        gamma_constraint = assert_args_is_list(gamma_constraint)
        # Dropout
        use_dropout = assert_args_is_list(use_dropout)
        rate = assert_args_is_list(rate)
        seed = assert_args_is_list(seed)
        if not isinstance(noise_shape, list):  # Special case, for shape, use tuple here.
            noise_shape = [noise_shape for _ in units]

        for x in [activation, kernel_regularizer, bias_regularizer, activity_regularizer, kernel_initializer,
                  bias_initializer, kernel_constraint, bias_constraint, use_bias, axis, momentum, epsilon,
                  center, scale, beta_initializer, gamma_initializer, moving_mean_initializer,
                  moving_variance_initializer, beta_regularizer, gamma_regularizer, beta_constraint,
                  gamma_constraint, rate, seed, noise_shape, use_dropout, use_normalization, normalization_technique]:
            if len(x) != len(units):
                raise ValueError("Error: Provide matching list of units", units, "and", x, "or simply a single value.")

        for x in normalization_technique:
            if x not in ["batch", "layer", "group", "instance"]:
                raise ValueError("ERROR Unknown normalization method, choose: batch, layer, group, instance, ...")

        # Deserialized args
        self._conf_units = list(units)
        self._conf_use_bias = list(use_bias)
        self._conf_activation = list([tf.keras.activations.get(x) for x in activation])
        self._conf_kernel_regularizer = list([tf.keras.regularizers.get(x) for x in kernel_regularizer])
        self._conf_bias_regularizer = list([tf.keras.regularizers.get(x) for x in bias_regularizer])
        self._conf_activity_regularizer = list([tf.keras.regularizers.get(x) for x in activity_regularizer])
        self._conf_kernel_initializer = list([tf.keras.initializers.get(x) for x in kernel_initializer])
        self._conf_bias_initializer = list([tf.keras.initializers.get(x) for x in bias_initializer])
        self._conf_kernel_constraint = list([tf.keras.constraints.get(x) for x in kernel_constraint])
        self._conf_bias_constraint = list([tf.keras.constraints.get(x) for x in bias_constraint])
        # Serialized args for norm
        self._conf_use_normalization = list(use_normalization)
        self._conf_normalization_technique = list(normalization_technique)
        self._conf_axis = list(axis)
        self._conf_momentum = list(momentum)
        self._conf_epsilon = list(epsilon)
        self._conf_center = list(center)
        self._conf_scale = list(scale)
        self._conf_beta_initializer = list([tf.keras.initializers.get(x) for x in beta_initializer])
        self._conf_gamma_initializer = list([tf.keras.initializers.get(x) for x in gamma_initializer])
        self._conf_moving_mean_initializer = list([tf.keras.initializers.get(x) for x in moving_mean_initializer])
        self._conf_moving_variance_initializer = list(
            [tf.keras.initializers.get(x) for x in moving_variance_initializer])
        self._conf_beta_regularizer = list([tf.keras.regularizers.get(x) for x in beta_regularizer])
        self._conf_gamma_regularizer = list([tf.keras.regularizers.get(x) for x in gamma_regularizer])
        self._conf_beta_constraint = list([tf.keras.constraints.get(x) for x in beta_constraint])
        self._conf_gamma_constraint = list([tf.keras.constraints.get(x) for x in gamma_constraint])
        # Dropout
        self._conf_use_dropout = list(use_dropout)
        self._conf_rate = list(rate)
        self._conf_seed = list(seed)
        self._conf_noise_shape = list(noise_shape)

        # Processed arguments for each layer.
        self._conf_mlp_dense_layer_kwargs = [{"units": self._conf_units[i],
                                              "use_bias": self._conf_use_bias[i],
                                              "name": self.name + "_dense_" + str(i),
                                              "activation": "linear",
                                              "activity_regularizer": None,
                                              "kernel_regularizer": self._conf_kernel_regularizer[i],
                                              "bias_regularizer": self._conf_bias_regularizer[i],
                                              "kernel_initializer": self._conf_kernel_initializer[i],
                                              "bias_initializer": self._conf_bias_initializer[i],
                                              "kernel_constraint": self._conf_kernel_constraint[i],
                                              "bias_constraint": self._conf_bias_constraint[i]}
                                             for i in range(self._depth)]
        self._conf_mlp_activ_layer_kwargs = [{"activation": self._conf_activation[i],
                                              "name": self.name + "_act_" + str(i),
                                              "activity_regularizer": self._conf_activity_regularizer[i]}
                                             for i in range(self._depth)]
        self._conf_mlp_norm_layer_kwargs = [{"name": self.name + '_norm_' + str(i),
                                             "axis": self._conf_axis[i],
                                             "epsilon": self._conf_epsilon[i],
                                             "center": self._conf_center[i],
                                             "scale": self._conf_scale[i],
                                             "beta_initializer": self._conf_beta_initializer[i],
                                             "gamma_initializer": self._conf_gamma_initializer[i],
                                             "beta_regularizer": self._conf_beta_regularizer[i],
                                             "gamma_regularizer": self._conf_gamma_regularizer[i],
                                             "beta_constraint": self._conf_beta_constraint[i],
                                             "gamma_constraint": self._conf_gamma_constraint[i]}
                                            for i in range(self._depth)]
        self._conf_mlp_drop_layer_kwargs = [{"name": self.name + '_drop_' + str(i),
                                             "rate": self._conf_rate[i],
                                             "noise_shape": self._conf_noise_shape[i],
                                             "seed": self._conf_seed[i]}
                                            for i in range(self._depth)]

    def build(self, input_shape):
        """Build layer."""
        super(MLPBase, self).build(input_shape)

    def get_config(self):
        """Update config."""
        config = super(MLPBase, self).get_config()
        config.update({
            # Dense
            "units": self._conf_units,
            'use_bias': self._conf_use_bias,
            'activation': [tf.keras.activations.serialize(x) for x in self._conf_activation],
            'activity_regularizer': [tf.keras.regularizers.serialize(x) for x in
                                     self._conf_activity_regularizer],
            'kernel_regularizer': [tf.keras.regularizers.serialize(x) for x in self._conf_kernel_regularizer],
            'bias_regularizer': [tf.keras.regularizers.serialize(x) for x in self._conf_bias_regularizer],
            "kernel_initializer": [tf.keras.initializers.serialize(x) for x in self._conf_kernel_initializer],
            "bias_initializer": [tf.keras.initializers.serialize(x) for x in self._conf_bias_initializer],
            "kernel_constraint": [tf.keras.constraints.serialize(x) for x in self._conf_kernel_constraint],
            "bias_constraint": [tf.keras.constraints.serialize(x) for x in self._conf_bias_constraint],
            # Norm
            "normalization_technique": self._conf_normalization_technique,
            "use_normalization": self._conf_use_normalization,
            "axis": list(self._conf_axis),
            "momentum": self._conf_momentum,
            "epsilon": self._conf_epsilon,
            "center": self._conf_center,
            "scale": self._conf_scale,
            "beta_initializer": [tf.keras.initializers.serialize(x) for x in self._conf_beta_initializer],
            "gamma_initializer": [tf.keras.initializers.serialize(x) for x in self._conf_gamma_initializer],
            "moving_mean_initializer": [tf.keras.initializers.serialize(x) for x in
                                        self._conf_moving_mean_initializer],
            "moving_variance_initializer": [tf.keras.initializers.serialize(x) for x in
                                            self._conf_moving_variance_initializer],
            "beta_regularizer": [tf.keras.regularizers.serialize(x) for x in self._conf_beta_regularizer],
            "gamma_regularizer": [tf.keras.regularizers.serialize(x) for x in self._conf_gamma_regularizer],
            "beta_constraint": [tf.keras.constraints.serialize(x) for x in self._conf_beta_constraint],
            "gamma_constraint": [tf.keras.constraints.serialize(x) for x in self._conf_gamma_constraint],
            # Dropout
            "use_dropout": self._conf_use_dropout,
            'rate': self._conf_rate,
            'noise_shape': self._conf_noise_shape,
            'seed': self._conf_seed
        })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='GraphMLP')
class GraphMLP(MLPBase):
    r"""Multilayer perceptron that consist of N dense layers for ragged embedding tensors.
    See layer arguments of :obj:`MLPBase` for configuration.
    This layer adds normalization for embeddings tensors of node or edge embeddings represented by a ragged tensor.
    The definition of graph, batch or layer normalization is different from the standard keras definition.
    See :obj:`kgcnn.layers.norm` for more information.
    """

    def __init__(self, units, **kwargs):
        """Initialize MLP as for dense."""
        super(GraphMLP, self).__init__(units=units, **kwargs)

        self.mlp_dense_layer_list = [DenseEmbedding(
            **self._conf_mlp_dense_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_activation_layer_list = [ActivationEmbedding(
            **self._conf_mlp_activ_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_dropout_layer_list = [
            DropoutEmbedding(**self._conf_mlp_drop_layer_kwargs[i]) if self._conf_use_dropout[i] else None for i in
            range(self._depth)]

        self.mlp_norm_layer_list = [None] * self._depth
        for i in range(self._depth):
            if self._conf_use_normalization[i]:
                if self._conf_normalization_technique[i] in ["batch", "BatchNormalization", "GraphBatchNormalization"]:
                    self.mlp_norm_layer_list[i] = GraphBatchNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i],
                        momentum=self._conf_momentum[i],
                        moving_mean_initializer=self._conf_moving_mean_initializer[i],
                        moving_variance_initializer=self._conf_moving_variance_initializer[i])
                elif self._conf_normalization_technique[i] in ["layer", "LayerNormalization",
                                                               "GraphLayerNormalization"]:
                    self.mlp_norm_layer_list[i] = GraphLayerNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i])
                else:
                    raise NotImplementedError(
                        "ERROR: Normalization via %s not supported." % self._conf_normalization_technique[i])

    def build(self, input_shape):
        """Build layer."""
        super(GraphMLP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.RaggedTensor): Input tensor with last dimension not None.

        Returns:
            tf.Tensor: MLP pass.
        """
        x = inputs
        for i in range(len(self._conf_units)):
            x = self.mlp_dense_layer_list[i](x, **kwargs)
            if self._conf_use_dropout[i]:
                x = self.mlp_dropout_layer_list[i](x, **kwargs)
            if self._conf_use_normalization[i]:
                x = self.mlp_norm_layer_list[i](x, **kwargs)
            x = self.mlp_activation_layer_list[i](x, **kwargs)
        out = x
        return out

    def get_config(self):
        """Update config."""
        config = super(GraphMLP, self).get_config()
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='MLP')
class MLP(MLPBase):
    r"""Multilayer perceptron that consist of N dense layers for ragged embedding tensors.
    See layer arguments of :obj:`MLPBase` for configuration.
    This layer adds normalization and dropout for normal tensor input.
    """

    def __init__(self, units, **kwargs):
        """Initialize MLP. See MLPBase."""
        super(MLP, self).__init__(units=units, **kwargs)

        self.mlp_dense_layer_list = [tf.keras.layers.Dense(
            **self._conf_mlp_dense_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_activation_layer_list = [tf.keras.layers.Activation(
            **self._conf_mlp_activ_layer_kwargs[i]) for i in range(self._depth)]

        self.mlp_dropout_layer_list = [
            tf.keras.layers.Dropout(**self._conf_mlp_drop_layer_kwargs[i]) if self._conf_use_dropout[i] else None for i
            in range(self._depth)]

        self.mlp_norm_layer_list = [None] * self._depth
        for i in range(self._depth):
            if self._conf_use_normalization[i]:
                if self._conf_normalization_technique[i] in ["batch", "BatchNormalization", "GraphBatchNormalization"]:
                    self.mlp_norm_layer_list[i] = tf.keras.layers.BatchNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i],
                        momentum=self._conf_momentum[i],
                        moving_mean_initializer=self._conf_moving_mean_initializer[i],
                        moving_variance_initializer=self._conf_moving_variance_initializer[i])
                elif self._conf_normalization_technique[i] in ["layer", "LayerNormalization",
                                                               "GraphLayerNormalization"]:
                    self.mlp_norm_layer_list[i] = tf.keras.layers.LayerNormalization(
                        **self._conf_mlp_norm_layer_kwargs[i])
                else:
                    raise NotImplementedError(
                        "ERROR: Normalization via %s not supported." % self._conf_normalization_technique[i])

    def build(self, input_shape):
        """Build layer."""
        super(MLP, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Forward pass.

        Args:
            inputs (tf.Tensor): Input tensor with last dimension not None.

        Returns:
            tf.Tensor: MLP pass.
        """
        x = inputs
        for i in range(len(self._conf_units)):
            x = self.mlp_dense_layer_list[i](x, **kwargs)
            if self._conf_use_dropout[i]:
                x = self.mlp_dropout_layer_list[i](x, **kwargs)
            if self._conf_use_normalization[i]:
                x = self.mlp_norm_layer_list[i](x, **kwargs)
            x = self.mlp_activation_layer_list[i](x, **kwargs)
        out = x
        return out

    def get_config(self):
        """Update config."""
        config = super(MLP, self).get_config()
        return config



import tensorflow.keras as ks

from kgcnn.layers.casting import ChangeTensorType
from kgcnn.layers.conv.gin_conv import GIN
from kgcnn.layers.modules import DenseEmbedding, OptionalInputEmbedding
from kgcnn.layers.mlp import GraphMLP, MLP
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.utils.models import update_model_kwargs

# How Powerful are Graph Neural Networks?
# Keyulu Xu, Weihua Hu, Jure Leskovec, Stefanie Jegelka
# https://arxiv.org/abs/1810.00826

model_default = {'name': "GIN",
                 'inputs': [{'shape': (None,), 'name': "node_attributes", 'dtype': 'float32', 'ragged': True},
                            {'shape': (None, 2), 'name': "edge_indices", 'dtype': 'int64', 'ragged': True}],
                 'input_embedding': {"node": {"input_dim": 95, "output_dim": 64}},
                 'gin_args': {"units": [64, 64], "use_bias": True, "activation": ['relu', 'linear'],
                              "use_normalization": True, "normalization_technique": "batch"},
                 'depth': 3, "dropout": 0.0, 'verbose': 10,
                 'last_mlp': {"use_bias": [True, True, True], "units": [64, 64, 64],
                              "activation": ['relu', 'relu', 'linear']},
                 'output_embedding': 'graph',
                 'output_mlp': {"use_bias": True, "units": 1,
                                "activation": "softmax"}
                 }


@update_model_kwargs(model_default)
def make_model(inputs=None,
               input_embedding=None,
               depth=None,
               gin_args=None,
               last_mlp=None,
               dropout=None,
               name=None,
               verbose=None,
               output_embedding=None,
               output_mlp=None
               ):
    """Make GIN graph network via functional API. Default parameters can be found in :obj:`model_default`.

    Args:
        inputs (list): List of dictionaries unpacked in :obj:`tf.keras.layers.Input`. Order must match model definition.
        input_embedding (dict): Dictionary of embedding arguments for nodes etc. unpacked in `Embedding` layers.
        depth (int): Number of graph embedding units or depth of the network.
        gin_args (dict): Dictionary of layer arguments unpacked in `GIN` convolutional layer.
        last_mlp (dict): Dictionary of layer arguments unpacked in last `MLP` layer before output or pooling.
        dropout (float): Dropout to use.
        name (str): Name of the model.
        verbose (int): Level of print output.
        output_embedding (str): Main embedding task for graph network. Either "node", ("edge") or "graph".
        output_mlp (dict): Dictionary of layer arguments unpacked in the final classification `MLP` layer block.
            Defines number of model outputs and activation.

    Returns:
        tf.keras.models.Model
    """

    # Make input
    assert len(inputs) == 2
    node_input = ks.layers.Input(**inputs[0])
    edge_index_input = ks.layers.Input(**inputs[1])

    # Embedding, if no feature dimension
    n = OptionalInputEmbedding(**input_embedding['node'],
                               use_embedding=len(inputs[0]['shape']) < 2)(node_input)
    edi = edge_index_input

    # Model
    # Map to the required number of units.
    n = DenseEmbedding(gin_args["units"][0], use_bias=True, activation='linear')(n)
    # n = MLP(gin_args["units"], use_bias=gin_args["use_bias"], activation=gin_args["activation"])(n)
    list_embeddings = [n]
    for i in range(0, depth):
        n = GIN()([n, edi])
        n = GraphMLP(**gin_args)(n)
        list_embeddings.append(n)

    # Output embedding choice
    if output_embedding == "graph":
        out = [PoolingNodes()(x) for x in list_embeddings]  # will return tensor
        out = [MLP(**last_mlp)(x) for x in out]
        out = [ks.layers.Dropout(dropout)(x) for x in out]
        out = ks.layers.Add()(out)
        out = MLP(**output_mlp)(out)
    elif output_embedding == "node":  # Node labeling
        out = n
        out = GraphMLP(**last_mlp)(out)
        out = GraphMLP(**output_mlp)(out)
        # no ragged for distribution supported atm
        out = ChangeTensorType(input_tensor_type='ragged', output_tensor_type="tensor")(out)
    else:
        raise ValueError("Unsupported graph embedding for mode `GIN`")

    model = ks.models.Model(inputs=[node_input, edge_index_input], outputs=out)
    return model
